// Person.java
//
// ICS 22 / CSE 22 Fall 2012
// Code Example
//
// There are two ways to check for equality of things in Java: == and the
// equals() method.  Which one you use depends on which question you're
// trying to answer and what types of things you're comparing.
//
// * The == operator always answers the question "Do these two variables
//   have the same value?"  In the case of two values of primitive types
//   (like, say, ints), == will tell you whether the two ints are the
//   same.  So, for example, this code will print "Woo-hoo!".
//
//       int i = 3;
//       int j = 3;
//       if (i == j)
//           System.out.println("Woo-hoo!");
//       else
//           System.out.println("Doh!");
//
//   In the case of two references -- which tell you where other objects
//   live, but are not themselves objects -- their values are the same
//   if they point to the same place; in other words, == tells you whether
//   they both refer to the same object.
//
//       Student s = new Student();
//       Student t = s;   // point t to where s points
//       if (s == t)
//           System.out.println("Woo-hoo!");   // this will be the output
//       else
//           System.out.println("Doh!");
//
//   This leads to some surprising behavior, though, when you're talking
//   about Strings:
//
//       String s1 = "Alex";
//       String s2 = "Al";
//       String s3 = s2 + "ex";
//       if (s1 == s3)
//           System.out.println("Woo-hoo!");
//       else
//           System.out.println("Doh!");     // this will be the output
//
//   If you found the String example a bit counterintuitive, you'll have to
//   adjust your intuition a bit: s1 and s3 aren't the same String object;
//   they're two different String objects that happen to have the same
//   sequence of characters in them.
//
// * When you're comparing objects for equality, though, what you're often
//   interested in is whether the two objects are equivalent (i.e., they
//   have the same "meaning").  For example, when you compare two Strings,
//   you generally want to know whether they have the same sequence of
//   characters in them (which is what it means for two Strings to "mean"
//   the same thing, i.e., to be equivalent), regardless of whether they're
//   the same object or not.  When you're interested in checking for object
//   equivalence, you call the equals() method on one of them -- effectively,
//   you're asking one of the objects "Hey, are you equivalent to this other
//   object?"  You're almost always going to want to use equals() to compare
//   two Strings, which you can do like this:
//
//       String s1 = "Alex";
//       String s2 = "Al";
//       String s3 = s2 + "ex";
//
//       // s1 and s3 now refer to different String objects that both happen
//       // to have the same sequence of characters in them
//
//       if (s1.equals(s3))
//           System.out.println("Woo-hoo!");  // this will be the output
//       else
//           System.out.println("Doh!");
//
// All classes have an equals() method by default.  The default version
// actually does the same thing that == does: it checks for whether the
// two objects live in the same location.  But you can override equals()
// with a new version in any of your classes, in effect telling Java
// "Here's how you decide whether two objects of this type are equivalent."
//
// An example implementation of the equals() method follows.  We have a
// class called Person, where Persons are defined with an ID number and
// a name.  We'll say that two Persons are equivalent if they have the
// same ID number and the same name.


public class Person
{
	private int id;
	private String name;
	
	
	public Person(int id, String name)
	{
		this.id = id;
		this.name = name;
	}
	
	
	public int getId()
	{
		return id;
	}
	
	
	public String getName()
	{
		return name;
	}
	
	
	// This is the implementation of the equals() method.  It must
	// always have this signature: it takes an Object reference as a
	// parameter.  Object references are special in Java; they can
	// point to any kind of Object.
	public boolean equals(Object x)
	{
		// Since Object references can point to anything, the compiler
		// can't know what x may be pointing to; it may be pointing to
		// a Person or it may not.  So we'll need to first check what
		// kind of object x is pointing to.  If it's pointing to a
		// Person, we'll compare the ids and names.  If it's not, we'll
		// just return false, since a Person is never equivalent to
		// some other kind of object.
		//
		// The "instanceof" operator in Java is how you ask what kind
		// of object a reference points to.  Here, I'm asking "Does x
		// point to a Person or to something else?"
		
		if (x instanceof Person)
		{
			// This is called a "typecast" or a "cast."  I have to do
			// this because the compiler will only ever assume that x
			// points to some kind of object, but will never assume what
			// kind of object it is.  So I ask for a new reference
			// pointing to the same object but with the type Person instead.
			//
			// Java will check this cast at runtime.  If x really points
			// to some other type, the cast will cause an exception to
			// be thrown.  I'm sure that will never happen, though,
			// because if I get to this line of code, x has already been
			// checked and found to point to a Person.
			
			Person p = (Person) x;


			// Here, I'm comparing the IDs and the names, returning true
			// if they're both the same and false otherwise.  Note that
			// I'm using == to compare the IDs, since they're ints (which
			// are primitive types), and equals() to compare the names,
			// since they're Strings (and I'm interested in comparing
			// the contents of the Strings, rather than whether the
			// Strings are really the same object).

			return id == p.id && name.equals(p.name);
		}
		else
		{
			// If x points to something other than a Person, I'll just
			// return false, because a Person is never equivalent to
            // something that isn't also a Person.
			
			return false;
		}
	}
}
